package sun.misc;

import java.util.TreeSet;
import java.util.SortedSet;
import java.security.AccessController;
import java.security.PrivilegedAction;

public class GC
{
    private static final long NO_TARGET = Long.MAX_VALUE;
    private static long latencyTarget;
    private static Thread daemon;
    private static Object lock;
    
    private GC() {
    }
    
    public static native long maxObjectInspectionAge();
    
    private static void setLatencyTarget(long latencyTarget) {
        GC.latencyTarget = latencyTarget;
        if (GC.daemon == null) {
            GC.Daemon.create();
        }
        else {
            GC.lock.notify();
        }
    }
    
    public static GC.LatencyRequest requestLatency(long n) {
        return new LatencyRequest(n);
    }
    
    public static long currentLatencyTarget() {
        long latencyTarget = GC.latencyTarget;
        return (latencyTarget == Long.MAX_VALUE) ? 0L : latencyTarget;
    }
    
    static void access$500(long latencyTarget) {
        setLatencyTarget(latencyTarget);
    }
    
    static {
        GC.latencyTarget = Long.MAX_VALUE;
        GC.daemon = null;
        GC.lock = new GC.LatencyLock();
    }
    
    private static class LatencyLock
    {
    }
    
    private static class Daemon extends Thread
    {
        @Override
        public void run() {
            while (true) {
                synchronized (GC.lock) {
                    long access$200 = GC.latencyTarget;
                    if (access$200 == Long.MAX_VALUE) {
                        GC.daemon = null;
                        return;
                    }
                    long maxObjectInspectionAge = GC.maxObjectInspectionAge();
                    if (maxObjectInspectionAge >= access$200) {
                        System.gc();
                        maxObjectInspectionAge = 0L;
                    }
                    try {
                        GC.lock.wait(access$200 - maxObjectInspectionAge);
                    }
                    catch (InterruptedException ex) {
                        continue;
                    }
                }
            }
        }
        
        private Daemon(ThreadGroup group) {
            super(group, "GC Daemon");
        }
        
        public static void create() {
            AccessController.doPrivileged(new PrivilegedAction<Void>() {
                @Override
                public Void run() {
                    ThreadGroup threadGroup;
                    for (threadGroup = (threadGroup = Thread.currentThread().getThreadGroup()); threadGroup != null; , threadGroup = threadGroup.getParent()) {}
                    GC.Daemon o = new Daemon(threadGroup);
                    o.setDaemon(true);
                    o.setPriority(2);
                    o.start();
                    GC.daemon = o;
                    return null;
                }
            });
        }
    }
    
    public static class LatencyRequest implements Comparable<GC.LatencyRequest>
    {
        private static long counter;
        private static SortedSet<GC.LatencyRequest> requests;
        private long latency;
        private long id;
        
        private static void adjustLatencyIfNeeded() {
            if (GC.LatencyRequest.requests == null || GC.LatencyRequest.requests.isEmpty()) {
                if (GC.latencyTarget != Long.MAX_VALUE) {
                    GC.access$500(Long.MAX_VALUE);
                }
            }
            else {
                GC.LatencyRequest latencyRequest = (GC.LatencyRequest)GC.LatencyRequest.requests.first();
                if (latencyRequest.latency != GC.latencyTarget) {
                    GC.access$500(latencyRequest.latency);
                }
            }
        }
        
        private LatencyRequest(long n) {
            if (n <= 0L) {
                throw new IllegalArgumentException("Non-positive latency: " + n);
            }
            this.latency = n;
            synchronized (GC.lock) {
                this.id = ++GC.LatencyRequest.counter;
                if (GC.LatencyRequest.requests == null) {
                    GC.LatencyRequest.requests = new TreeSet();
                }
                GC.LatencyRequest.requests.add(this);
                adjustLatencyIfNeeded();
            }
            return;
        }
        
        public void cancel() {
            synchronized (GC.lock) {
                if (this.latency == Long.MAX_VALUE) {
                    throw new IllegalStateException("Request already cancelled");
                }
                if (!GC.LatencyRequest.requests.remove(this)) {
                    throw new InternalError("Latency request " + this + " not found");
                }
                if (GC.LatencyRequest.requests.isEmpty()) {
                    GC.LatencyRequest.requests = null;
                }
                this.latency = Long.MAX_VALUE;
                adjustLatencyIfNeeded();
            }
        }
        
        @Override
        public int compareTo(GC.LatencyRequest latencyRequest) {
            long n = this.latency - latencyRequest.latency;
            if (n == 0L) {
                n = this.id - latencyRequest.id;
            }
            return (n < 0L) ? -1 : (n > 0L);
        }
        
        @Override
        public String toString() {
            return GC.LatencyRequest.class.getName() + "[" + this.latency + "," + this.id + "]";
        }
        
        static {
            GC.LatencyRequest.counter = 0L;
            GC.LatencyRequest.requests = null;
        }
    }
}
